/*=====================================================================================================================
 Arduino-Stationsuhr by Ingo Strecker, DL6IS

 2014-10-26: Temperaturoffsets im Konfigurationsbereich
 2014-03-29: Blinkende Ziffern beim Einstellen von Counter / Alarm
             Bis Anzeige SET bei Alarm erscheint, nicht dauernd Ein, Aus, Ein, Aus, ....
 2014-03-23: Rufzeichenanzeige hinzu
             Anzeige von MEZ, bzw. MESZ, und UTC (anstelle von SZ und WZ)
 2014-03-08: Konfigurationsbereich eingerichtet
             Wahl der Temperatursensoren möglich - Nur der/die gewählte(n) Sensor(en) werden auf dem Display angezeigt.
             Sommer-/Winterzeit bei UTC nicht anzeigen

=======================================================================================================================*/

//#####################################################################################################################
//Konfigurationsbereich. Hier dürfen Parameter angepasst werden. Und nur hier!

#define Expander 0x20        // Adresse des Portexpanders: PCF8574 = 0x20, PCF8574A = 0x38

char Rufzeichen[] = "DL6IS"; // wird links oben angezeigt, wenn AussenSensorDa = false

bool AussenSensorDa = true;  // wenn true (Sensor vorhanden), dann muss folgende Adresse angegeben werden!
byte addrAussen[8] = {0x28, 0xFF, 0x3C, 0x5F, 0x2C, 0x4, 0x0, 0xBB};   //Adresse DS18B20

bool InnenSensorDa = true;   // wenn true (Sensor vorhanden), dann muss folgende Adresse angegeben werden!
byte addrInnen[8] = {0x28, 0x20, 0x26, 0x29, 0x5, 0x0, 0x0, 0x1A}; //Adresse DS18B20

unsigned int backlightTimeOn = 3000;  // Soviele Millisekunden bleibt das Backlight bei kurzem Tastendruck an

float OffsetInnen = 1.0;     // Dieser Wert in Grad Celsius wird von der Messtemperatur abgezogen
float OffsetAussen = 0.5;    // Dieser Wert in Grad Celsius wird von der Messtemperatur abgezogen

//Ende des Konfigurationsbereichs.
//#####################################################################################################################*/


#define NO_PORTB_PINCHANGES // Port B wird nicht benützt für pin change interrupts
#define NO_PORTC_PINCHANGES // Port C wird nicht benützt für pin change interrupts
#include <PinChangeInt.h>   // pin change interrupts
#include <AdaEncoder.h>     // Drehgeber
#include <EEPROM.h>
#include <DCF77.h>
#include <Time.h>
#include <Wire.h>
#include <DS1307RTC.h>
#include <OneWire.h>
#include <glcd.h>
#include "fonts/SystemFont5x7.h"       // system font
#include "fonts/fixed5x11.h"           // fixed width font - numerals only
#include "fonts/fixednums10x22.h"      // fixed width font - numerals only http://www.stefan-buchgeher.info/elektronik/glcd/glcd_kap7.html#Kap7
#include "bitmaps/DCFSymbol.h"         // Symbol für DCF-Empfang

#define PIN0 0               // Pin für Drehencoder, aber auch serielle Schnittstelle
#define PIN1 1               // Pin für Drehencoder, aber auch serielle Schnittstelle
#define DCF_PIN 2	           // DCF77 Daten
#define DS_PIN 3             // OneWire Pin
#define DCF_INTERRUPT 0		   // DCF77 Interrupt
bool dcfSync = false;        // Zeit empfangen

time_t prevDisplay = 0;      // vergangene Zeit
int memDST = 0;              // EEPROM-Adresse - Sommer-/Winterzeit
int memAlarmHour = 1;        //       "        - Inhalt Alarmstunde
int memAlarmMinute = 2;      //       "        - Inhalt Alarmminute
int Sommerzeit;              // Inhalt Countdown
byte data[12];
//bool InnenSensor = true;
unsigned int WaitForSensor;
float celsius;
byte oneSecond = 5;
byte localTime = true;
byte button = 0b10111111;    // Backlight beim Start einschalten (Bit7 = 0, wenn nicht gewünscht dann Bit7 = 1)
byte backlightOn = true;     // Backlight beim Start einschalten (false, wenn nicht gewünscht)
unsigned long blink = 500;
unsigned long buttonPressedLight, buttonPressedCounter, buttonPressedAlarm;
unsigned long backlightPeriod, blinkTime;
byte buttonDown;
uint8_t encoderPos = 0;
int8_t clicks=0;
char id=0;
byte Counter = 0;             // Inhalt Countdown
bool counterSet = false;      // Countdown stellen
bool startCounter = false;    // Countdown gestartet
bool checkCounter = true;     // Piepsen Countdown gestoppt
time_t counterStart;          // Uhrzeit, wo Countdown gestartet wird
byte alarmHour = 0;           // Inhalt Alarmstunde
byte alarmMinute = 0;         // Inhalt Alarmminute
bool alarmSet = false;        // Alarm stellen
bool startAlarm = false;      // Alarm gestartet
bool checkAlarm = true;       // Piepsen Alarm gestoppt
byte encoderPosOld = 0;
bool setHour = true;          // wird Alarmstunde gestellt
bool countMinutes = false;
bool countMinutesSet = false;
                                                                       

OneWire ds(DS_PIN);  // Datenpin DS1820 - Temperaturen (ein Pull-Up Widerstand wird benötigt; 10k)
DCF77 DCF = DCF77 (DCF_PIN, DCF_INTERRUPT);   // DCF77 Empfang

//###########################################################
void displayCounter(bool visible = true) {
  GLCD.SelectFont(fixed5x11);
  if (visible) {
    GLCD.GotoXY(0, 53);  GLCD.print(Counter);  GLCD.print("  ");
  } else {
    GLCD.GotoXY(0, 53);  GLCD.print("   ");
  }
}

//###########################################################
void displayAlarm(bool visible = true) {
  GLCD.SelectFont(fixed5x11);
  if (visible) {
    GLCD.GotoXY(162, 53);  GLCD.Printf("%02d:%02d", alarmHour, alarmMinute);
  } else {
    if (setHour) {
      GLCD.GotoXY(162, 53);  GLCD.Printf("  :%02d", alarmMinute);
    } else {
      GLCD.GotoXY(162, 53);  GLCD.Printf("%02d:  ", alarmHour);
    }
  }
}

//###########################################################
void blinkSettings(char Blinken[]) {
  if (millis() - blinkTime <= 250) {
    if (Blinken == "Count")
    {
      displayCounter(false);
    }
    if (Blinken == "Alarm")
    {
      displayAlarm(false);
    }
  }
  if (millis() - blinkTime > 250) {
    if (Blinken == "Count")
    {
      displayCounter();
    }
    if (Blinken == "Alarm")
    {
      displayAlarm();
    }
    if (millis() - blinkTime >=500) {
      blinkTime = millis();
    }
  }
}

//###########################################################
void setup() {
  //Serial.begin(115200);
  DCF.Start();
  pinMode(PIN0, INPUT);
  pinMode(PIN1, INPUT);
  AdaEncoder::addEncoder('a', PIN0, PIN1);
  Wire.begin();
  expanderWrite(Expander, button);  // Pin0-Pin7 Eingänge
  time_t rtcTime = RTC.get();
  if (rtcTime != 0) {
    setTime(rtcTime);
    setSyncInterval(10);
    setSyncProvider (RTC.get);
  }
  Sommerzeit = EEPROM.read(memDST);
  if (Sommerzeit != 1) Sommerzeit = 0;
  alarmHour = EEPROM.read(memAlarmHour);
  if (alarmHour == 255) alarmHour = 0;
  alarmMinute = EEPROM.read(memAlarmMinute);
  if (alarmMinute == 255) alarmMinute = 0;
  GLCD.Init();
  GLCD.SelectFont(System5x7);
  if (AussenSensorDa) {
	  GLCD.GotoXY(0, 0);  GLCD.print(F("Aussen:"));
  } else {
    GLCD.GotoXY(0, 0);  GLCD.print(Rufzeichen);
  }
  if (InnenSensorDa) {
	  GLCD.GotoXY(GLCD.Right - 35, 0);  GLCD.print(F("Innen:"));
  }
  GLCD.GotoXY(0, 35);  GLCD.print(F("CntSek"));
  GLCD.GotoXY(0, 44);  GLCD.print(F("Aus"));
  GLCD.GotoXY(162, 35);  GLCD.print(F("Alarm"));
  if (startAlarm) {
    GLCD.GotoXY(168, 44);  GLCD.print(F("Ein"));
    checkAlarm = false;
    startAlarm = true;
  } else {
    GLCD.GotoXY(168, 44);  GLCD.print(F("Aus"));
    checkAlarm = true;
    startAlarm = false;
  }    
  displayCounter(true);
  displayAlarm(true);
  ZeigeSommerWinterzeit(true);
}

//###########################################################
void loop() {
  CheckEncoder();
  Wire.requestFrom(Expander, 1);
  if (Wire.available()) {
    buttonDown = Wire.read(); 
    //---------------------------------------------------------------------------------------------------
    if (~buttonDown & 8) {   // Backlight - Taster 4 gedrückt; ~button = invertiert, da LOW wenn gedrückt
      button &= ~(1 << 6);    // Backlight einschalten - P6
      expanderWrite(Expander, button);
      delay (150);
      button |= (1 << 8);   // Taster 4 (vorher 2) nicht mehr gedrückt
      expanderWrite(Expander, button);
      backlightPeriod = millis();
      backlightOn = false;
      if (millis() - buttonPressedLight > 500) {
        backlightOn = true;
      }
      } else {
      buttonPressedLight = millis();
    }
    //---------------------------------------------------------------------------------------------------
    if (~buttonDown & 4) {   // UTC - Taster 3 gedrückt; ~button = invertiert, da LOW wenn gedrückt
      localTime = localTime ^ 1;
      Piepsen(1, 50);
      delay(50);
    }    
    //---------------------------------------------------------------------------------------------------
    if (~buttonDown & 2) {   // Countdown - Taster 2 gedrückt; ~button = invertiert, da LOW wenn gedrückt
      delay (150);
      Wire.requestFrom(Expander, 1);
      buttonDown = Wire.read();
      if (millis() - buttonPressedCounter > 3000 && !countMinutesSet) {
        countMinutes = countMinutes ^ 1;
        Piepsen(1, 30);
        GLCD.SelectFont(System5x7);
        if (countMinutes) {
          GLCD.GotoXY(0, 35);  GLCD.print(F("CntMin"));
        } else {
          GLCD.GotoXY(0, 35);  GLCD.print(F("CntSek"));
        }
        countMinutesSet = true;
      }
      if ((~buttonDown & 2)) {
        if (millis() - buttonPressedCounter > 500) {  // Counter stellen
          counterSet = true;
          startCounter = false;
          encoderPosOld = encoderPos;
          GLCD.SelectFont(System5x7);
          GLCD.GotoXY(0, 44);  GLCD.print(F("SET"));
        }
      } else {
        GLCD.SelectFont(System5x7);
        if (!checkCounter) {        // Piepsen ausschalten
          checkCounter = true;
          GLCD.GotoXY(0, 44);  GLCD.print(F("Aus"));
        } else {
          if (!counterSet) {        // Counter wird nicht gestellt, also Counter starten (und stoppen)
            if (Counter > 0) {
              GLCD.SelectFont(System5x7);
              counterStart = now();
              startCounter = startCounter ^ 1;    // Start / Stop Counter
              countMinutesSet = false;
              if (startCounter) {
                GLCD.GotoXY(0, 44);  GLCD.print(F("Ein"));
              } else {
                GLCD.GotoXY(0, 44);  GLCD.print(F("Aus"));
              }
            }              
          }
        }
      }
    } else {
      buttonPressedCounter = millis();
    }
    //---------------------------------------------------------------------------------------------------
    if (~buttonDown & 1) {   // Wecker - Taster 1 gedrückt; ~button = invertiert, da LOW wenn gedrückt
      delay (150);    // Entprellen
      GLCD.SelectFont(System5x7);
      Wire.requestFrom(Expander, 1);
      buttonDown = Wire.read();
      if ((~buttonDown & 1)) {
        if (millis() - buttonPressedAlarm > 500) {  // Wecker stellen
          alarmSet = true;
          startAlarm = false;
          encoderPosOld = encoderPos;
          GLCD.GotoXY(168, 44);  GLCD.print(F("SET"));
          delay(500);   // Wartezeit, damit nicht Aus statt SET angezeigt wird.
        } //else {
      } else {
        if (!checkAlarm) {           // Piepsen ausschalten
          checkAlarm = true;
          startAlarm = false;
          GLCD.GotoXY(168, 44);  GLCD.print(F("Aus"));
        } else {
          if (!alarmSet) {           // Wecker wird nicht gestellt, also Wecker einschalten
            checkAlarm = false;
            startAlarm = true;
            GLCD.GotoXY(168, 44);  GLCD.print(F("Ein"));
          }
        }
      }      
      //delay(100);
    } else {
      buttonPressedAlarm = millis();
    } 
    //---------------------------------------------------------------------------------------------------
    if (~buttonDown & 16) {   // Encoder - Taster 5 gedrückt; ~button = invertiert, da LOW wenn gedrückt
      GLCD.SelectFont(System5x7);
      delay(50);
      if (counterSet) {
        startCounter = false;    // Stop Counter
        counterSet = false;
        GLCD.GotoXY(0, 44);  GLCD.print(F("Aus"));
      }
      if (alarmSet) {
        if (!setHour) {
          alarmSet = false;
          setHour = true;
          //checkAlarm = true;    // Alarm sofort einschalten
          checkAlarm = false;     // Alarm sofort einschalten
          startAlarm = true;      // Alarm sofort einschalten
          GLCD.GotoXY(168, 44);  GLCD.print(F("Ein"));    // Alarm sofort einschalten
          //GLCD.GotoXY(168, 30);  GLCD.print(F("Aus"));  // Alarm sofort einschalten
          EEPROM.write(memAlarmHour, alarmHour);
          EEPROM.write(memAlarmMinute, alarmMinute);
        } else {
          if (setHour) {
            setHour = false;
          }
        }
      }
      displayCounter(true);
      displayAlarm(true);
      delay(100);
    }
  } 
  //---------------------------------------------------------------------------------------------------
  if (counterSet) {   // Countdown einstellen
    if (encoderPos != encoderPosOld) {
      if (encoderPos > encoderPosOld) {
        if (Counter <= 254) {
          Counter++;
        }
      } else {
        if (Counter > 0) {
          Counter--;
        }
      }
      encoderPosOld = encoderPos;
    }
    blinkSettings("Count");
  }
  if (alarmSet) {
    GLCD.SelectFont(fixed5x11);
    if (setHour) {    // Stunden stellen
      if (encoderPos != encoderPosOld) {
        if (encoderPos > encoderPosOld) {
          if (alarmHour <= 22) {
            alarmHour++;
          }            
        } else {
          if (alarmHour > 0) {
            alarmHour--;
          }
        }                      
        encoderPosOld = encoderPos;
      }
      blinkSettings("Alarm");
    } else {     // Minuten stellen
      if (encoderPos != encoderPosOld) {
        if (encoderPos > encoderPosOld) {
          if (alarmMinute <= 58) {
            alarmMinute++;
          }            
        } else {
          if (alarmMinute > 0) {
            alarmMinute--;
          }            
        }          
        encoderPosOld = encoderPos;
      }
      blinkSettings("Alarm");
    }
  }

  if (!dcfSync) {
    syncDCF77();      //#############################################
  }
  if (millis() - backlightPeriod > backlightTimeOn && !backlightOn) {
    button |= (1 << 6);   // Backlight ausschalten - P6
    expanderWrite(Expander, button);
    backlightOn = true;   // soll verhindern, dass die Schleife permanent durchlaufen wird, obwohl Backlight bereits aus ist.
  } 
  if (oneSecond == 5) {
/*    if (InnenSensor) {               // Sensoren abwechseln abfragen damit es funktioniert (hat aber auch schon mal gleichzeitig geklappt)
      if (InnenSensorDa) */
				StartTemperature(addrInnen);
/*    } else {
      if (AussenSensorDa) */
				StartTemperature(addrAussen);
//   } 
    oneSecond = 0;
    WaitForSensor = millis();
  }
  //-----------------------------------------------------------------------------------------------------------------------
  if(now() != prevDisplay) //update the display only if the time has changed
  {
    prevDisplay = now();
    //Serial.println(now());
    if (hour() == 2 && minute() == 0 && second() == 0) { // Nachts um 02:00 Uhr mit DCF77 syncen
    //if (minute() % 15 == 0 && second() == 0 && dcfSync) {  // alle 15 min. mit DCF77 syncen
      syncDCF77();
    }
	digitalClockDisplay();
    if (startCounter) {   // Countdown läuft.....
      if (!countMinutes) {
        Counter--;
      } else {
        if ((prevDisplay - counterStart) % 60 == 0) { // nicht sekündlich, sondern minütlich runter zählen!
          Counter--;
        }
      }        
      if (Counter == 0) {
        startCounter = false;   // Countdown fertig
        checkCounter = false;   // Piepsen...
        encoderPos = 1;
      }
    }
    displayCounter(true); 
    if (!checkCounter) {    // Countdown = 0; Piepsen bis checkCounter = true
      Piepsen(2, 50);
    }
    /*Serial.print(checkAlarm);  Serial.print("  ");  Serial.println(startAlarm);*/
    if (!checkAlarm && startAlarm && !alarmSet && hour(prevDisplay) == alarmHour && minute(prevDisplay) == alarmMinute) {
      Piepsen(3, 50);
    }
    
    GLCD.SelectFont(fixed5x11);
    //GLCD.GotoXY(0, 53);  GLCD.print(Counter);  GLCD.print("  ");
    //GLCD.GotoXY(162, 53);  GLCD.Printf("%02d:%02d", alarmHour, alarmMinute);
		if (AussenSensorDa) {
			float aussen = GetTemperature(addrAussen) - OffsetAussen;
			byte aussenSpaces = AnzahlSpaces(aussen);
			GLCD.GotoXY(0, 9);  GLCD.print(aussen, 1);  GLCD.print(F("`C"));
			for (int x = 0; x < 7 - aussenSpaces; x++) {   // höchstens 7 Stellen; mit Leerzeichen auffüllen, wenn weniger Stellen
				GLCD.print(" ");
			}
		}
		if (InnenSensorDa) {
			float innen = GetTemperature(addrInnen) - OffsetInnen;
			byte innenSpaces = AnzahlSpaces(innen);
			GLCD.GotoXY(150, 9);
			for (int x = 0; x < 7 - innenSpaces; x++) {   // höhstens 7 Stellen; mit Leerzeichen auffüllen, wenn weniger Stellen
				GLCD.print(" ");
			}
			GLCD.print(innen, 1);  GLCD.print(F("`C"));
		}
/*    InnenSensor = InnenSensor ^ 1; */
    oneSecond++; 
  }
}

//###########################################################
void syncDCF77() {
  button &= ~(1 << 5);   // DCF77 einschalten - P5
  expanderWrite(Expander, button);
  //DCF.Start();
  if (millis() - blink <= 500) {
    GLCD.DrawBitmap(DCFSymbol, 91, 0);
  } 
  if (millis() - blink > 500) {
    GLCD.FillRect(91, 0, 8, 8, WHITE);
    if (millis() - blink >=1000) {
      blink = millis();
    }
  }
  time_t DCFtime = DCF.getTime(); // Check if new DCF77 time is available
  if (DCFtime != 0) {
    RTC.set (DCFtime);
    setTime (DCFtime);
    dcfSync = true;
    //DCF.Stop();
    setSyncInterval(10);
    setSyncProvider (RTC.get);
    //GLCD.DrawBitmap(DCFSymbol, 0, GLCD.Bottom - 7);
    GLCD.DrawBitmap(DCFSymbol, 91, 0);
    Piepsen(1, 10);
    button |= (1 << 5);    // DCF77 ausschalten - P5
    expanderWrite(Expander, button);
    if (Sommerzeit != DCF.CEST) {
      Sommerzeit = DCF.CEST;
      EEPROM.write(memDST, Sommerzeit);
    }
    ZeigeSommerWinterzeit(true);
  } else dcfSync = false;
}

//###########################################################
void Piepsen(byte anzahl, byte pause) {
  for (int i=0; i<anzahl; i++) {
    button &= ~(1 << 7);   // Summer einschalten - P7
    expanderWrite(Expander, button);
    delay (pause);
    button |= (1 << 7);   // Summer ausschalten - P7
    expanderWrite(Expander, button);
    delay(pause);
  }
}

//###########################################################
void CheckEncoder() {
  encoder *thisEncoder;
  thisEncoder = AdaEncoder::genie(&clicks, &id);
  if (thisEncoder != NULL) {
    thisEncoder = AdaEncoder::getFirstEncoder();
    if (clicks > 0) {
      encoderPos += 1;
    }
    if (clicks < 0) {
      encoderPos -= 1;
    }
  }
}    

//###########################################################
byte AnzahlSpaces(float temp) {
  int spaces = int(log10(temp)) + 1 + 4;    // int(log10(innen)) + 1 = Anzahl Stellen der Zahl; + 4 = ".0°C"
  if (temp < 0) {    // Minusgrade
    spaces++;
  }
  return spaces;
}    

//###########################################################
byte expanderRead(int i2caddr) {
  int _data = -1;
  Wire.requestFrom(i2caddr, 1);
  if (Wire.available()) {
    _data = Wire.read();
  }
  return _data;
}

//###########################################################
void expanderWrite(int i2caddr, byte data)
{
  Wire.beginTransmission(i2caddr);
  Wire.write(data);
  Wire.endTransmission();
}

//###########################################################
void ZeigeSommerWinterzeit (bool anzeigen) {
  GLCD.SelectFont(System5x7);
  GLCD.GotoXY(GLCD.CenterX - (GLCD.StringWidth(" MEZ ") / 2), GLCD.Bottom - 6);
  if (anzeigen) {
    if (Sommerzeit == 1) {
      GLCD.GotoXY(GLCD.CenterX - (GLCD.StringWidth("MESZ") / 2), GLCD.Bottom - 6);
      GLCD.print(F("MESZ"));
    } else {
      GLCD.print(F(" MEZ "));
    }
  } else {
    GLCD.GotoXY(GLCD.CenterX - (GLCD.StringWidth(" UTC ") / 2), GLCD.Bottom - 6);
    GLCD.print(F(" UTC "));
  }
}

//###########################################################
void StartTemperature(byte* address) {
  ds.reset();
  //ds.skip();
  ds.select(address);
  ds.write(0x44, 0);        // start conversion, with not parasite power on at the end
//  while (ds.read() == 0) {  // reading timeslot (0 = not done with conversion)
//	  delay(100);             // and wait until done (not parasitically powered)
//  } 
}

//###########################################################
float GetTemperature(byte* address) {
  ds.reset();
  ds.select(address);
  ds.write(0xBE);         // Read Scratchpad
  for (int i = 0; i < 9; i++)           // we need 9 bytes
    data[i] = ds.read();
  int16_t raw = (data[1] << 8) | data[0];
  if (address[0] == 0x10) {  // DS18S20
	  raw = raw << 3; // 9 bit resolution default
	  if (data[7] == 0x10) {
		  // "count remain" gives full 12 bit resolution
		  raw = (raw & 0xFFF0) + 12 - data[6];
	  }
  } else {  // DS18B20, DS1822
    byte cfg = (data[4] & 0x60);
    // at lower res, the low bits are undefined, so let's zero them
    if (cfg == 0x00) raw = raw & ~7;  // 9 bit resolution, 93.75 ms
    else if (cfg == 0x20) raw = raw & ~3; // 10 bit res, 187.5 ms
    else if (cfg == 0x40) raw = raw & ~1; // 11 bit res, 375 ms
    //// default is 12 bit resolution, 750 ms conversion time
  }
  return (float)raw / 16.0;
}

//###########################################################
void digitalClockDisplay() {
  time_t t = now();
  if (localTime) {
    //t = now();
    GLCD.SetDisplayMode(NON_INVERTED);
    ZeigeSommerWinterzeit(true);
  } else {    
    // UTC
    int UTCTimeDifference = (Sommerzeit ? 2 : 1) * SECS_PER_HOUR;
    t -= UTCTimeDifference;
    //t = now() - UTCTimeDifference;
    GLCD.SetDisplayMode(INVERTED);
    ZeigeSommerWinterzeit(false);
  }    
  GLCD.SelectFont(fixednums10x22);
  GLCD.GotoXY(51, 13);
  GLCD.Printf("%02d:%02d:%02d", hour(t), minute(t), second(t));
  //GLCD.SelectFont(System5x7);
  GLCD.SelectFont(fixed5x11);
  GLCD.GotoXY(53, 40);
  GLCD.print(dayShortStr(weekday(t)));
  GLCD.print (" ");
  GLCD.Printf("%02d.%02d.%04d", day(t), month(t), year(t));
}
